#!/bin/bash

set -o nounset
set -o errexit

download()
{
	wget -nvc -q --show-progress --no-cache -P /tmp/redpack -T 15 $1
	if [ $? -ne 0 ]; then echo "Something went wrong when downloading a file at $1. Exiting..."; exit 1; fi
}

echo ""
echo "Welcome to your new Redclaw Linux system. If you see this, it is working fine."
echo ""
echo "You might have seen an error with XDG finding a directory that does not exist, but you can safely ignore that."
echo ""
echo "This is only the base system, but installation is almost complete."
echo ""
echo "This script will configure one network interface, needed to fetch and install the rest of the packages, plus the package manager \"redpack\"."
echo ""
echo "Finally, final configurations will finish the installation, requiring a reboot right after, which will be on you."
echo ""
echo "If you want to manually configure the system using the Configuration Guide, or you are satisfied with this base system, you may choose to move forward with this script."
echo ""
read -p "Would you like to continue? [Y/n]: "
if [ "$REPLY" = "n" -o "$REPLY" = "N" ]; then echo "Bye."; exit 0; fi

# configure networking

clear
echo ""
echo "A network connection is required to finish the installation."
echo ""
echo "Packages like ssh, lynx, redpack, and some systemd services rely on a network connection for most of their functionality."
echo ""
echo "This script will set up one adapter with a dynamic configuration. Refer to the Configuration Guide to set up multiple adapters, as well as setting a static configuration if preferred."
echo ""
echo "Select an interface from the list below:"
echo ""
ls /sys/class/net | grep -oE '[e|w].*'

LOOP=1
while [ $LOOP -eq 1 ]; do
	echo ""
	read -p "Interface: "
	for net in $(ls /sys/class/net | grep -oE '[e|w].*'); do
		if [ "$REPLY" = "$net" ]; then
			case $net in
				"e"*)
					echo -n " -- Creating interface file for systemd-networkd and systemd-resolved..."
					mv /etc/systemd/network/{00-interface.network.NOUSE,10-$net.network}
					sed "s/interface/$net/" -i /etc/systemd/network/10-$net.network
					echo " Done."
					echo "Ensure an Ethernet cable is connected and working."
					;;
				"w"*)
					echo -n " -- Creating interface file for systemd-networkd and systemd-resolved..."
					mv /etc/systemd/network/{00-interface.network.NOUSE,20-$net.network}
					sed "s/interface/$net/" -i /etc/systemd/network/20-$net.network
					echo " Done."
					if systemctl status wpa_supplicant@$net | grep -Eq "running"; then
						echo -n " -- Stopping wpa_supplicant..."
						systemctl stop wpa_supplicant@$net
						echo " Done."
					else
						echo -n " -- Bringing up \"$net\"..."
						ip link set $net up
						echo " Done."
					fi
					echo ""
					echo "We will set up a Wi-Fi connection using wpa_supplicant."
					echo "The passphrase will be masked when entered for privacy and security reasons."
					echo "If the network is open/public, leave the passphrase blank."
					echo "Refer to the Configuration Guide for setting up a hidden network."
					echo ""
					echo "Scanning for networks (wait a few seconds)..."
					echo ""
					iw wlo1 scan | grep 'SSID:' | sed 's/\tSSID: //' | sort -u
					echo ""
					read -p "SSID: " SSID
					read -sp "Passphrase: " PASSWORD
					echo ""
					echo ""
					echo -en " -- Creating \"/etc/wpa_supplicant/wpa_supplicant-$net.conf\"..."
					if [ -z "$PASSWORD" ]; then wpa_passphrase "$SSID" password | sed -e '/#/d' -e 's/psk=[0-9a-f]*/key_mgmt=NONE/' > /etc/wpa_supplicant/wpa_supplicant-$net.conf
					else wpa_passphrase "$SSID" "$PASSWORD" | sed '/#/d' > /etc/wpa_supplicant/wpa_supplicant-$net.conf
					fi
					echo " Done."
					echo -n " -- Enabling Systemd unit..."
					systemctl enable wpa_supplicant@$net
					echo " Done."
					echo -n " -- Starting wpa_supplicant..."
					systemctl start wpa_supplicant@$net
					echo " Done."
					;;
			esac
			echo ""
			echo "Date and time will automatically be synchronized once connected."
			echo ""
			LOOP=0
		fi
	done
done

echo -n " -- Restarting systemd-networkd and systemd-resolved..."
systemctl restart systemd-networkd
systemctl restart systemd-resolved
echo " Done."

# ping to ensure connection
echo ""
echo "Now pinging the main Redpack server (scot.butzbach.net) to test the connection."
echo "To stop the process, press Ctrl+C."
ATTEMPT=1
while [ $ATTEMPT -ge 1 ]; do
	echo -n " -- Ping attempt $ATTEMPT..."
	if ! ping -c 1 scot.butzbach.net 1>/dev/null 2>&1; then echo " FAIL"; else echo " SUCCESS"; break; fi
	sleep 1
	ATTEMPT=$((ATTEMPT + 1))
done

# fetch certificates
echo ""
echo " -- Fetching certificates..."
update-ca-certificates

# ask if documentation will installed with others
INSTALLDOCS=0
echo ""
read -p "Would you like to install documentation alongside the other packages? [y/N]: "
if [ "$REPLY" = "y" -o "$REPLY" = "Y" ]; then INSTALLDOCS=1; fi

# prepare privilege separation
clear
echo " -- Preparing privilege separation..."
groupadd -g 17 atd
groupadd -g 22 fcron
groupadd -fg 27 polkitd
groupadd -g 31 exim
groupadd -g 50 sshd
useradd -c "atd Daemon User" -d /dev/null -g atd -s /usr/bin/false -u 17 atd 2>/dev/null
useradd -c "fcron Daemon User" -d /dev/null -g fcron -s /usr/bin/false -u 22 fcron 2>/dev/null
useradd -c "PolicyKit Daemon User" -d /etc/polkit-1 -g polkitd -s /usr/bin/false -u 27 polkitd 2>/dev/null
useradd -c "exim Daemon User" -d /dev/null -g exim -s /usr/bin/false -u 31 exim 2>/dev/null
useradd -c "sshd Daemon User" -d /var/lib/sshd -g sshd -s /usr/bin/false -u 50 sshd 2>/dev/null

echo " -- Preparing directories..."
install -dm700 /var/lib/sshd -g sys
install -dm750 /var/spool/exim -o exim -g exim
chmod a+wt /var/mail

echo ""
echo " -- Fetching Redpack data files..."
mkdir /tmp/redpack
for x in {{cdn,hash}.master,urls.p{erl,ython}}.redpack; do
	download https://scot.butzbach.net/redpack/d/$x
	mv /{tmp,var/lib}/redpack/$x
done
if [ $INSTALLDOCS -eq 1 ]; then
	cat /var/lib/redpack/list.{installed,release.doc}.redpack | sort > /tmp/redpack/list.installed.redpack
	mv /{tmp,var/lib}/redpack/list.installed.redpack
fi
rm /var/lib/redpack/list.release.doc.redpack

# install release packages
echo ""
echo " -- Fetching packages..."
pushd /tmp/redpack > /dev/null
while read pkg; do
	case $pkg in
		perl-[a-z]*)
			url=$(grep -E "^$pkg\|" /var/lib/redpack/urls.perl.redpack | awk -F '|' '{print $2}')
			download $url
			download https://scot.butzbach.net/redpack/d/$pkg-install.sh
			;;
		python-doc-*) download https://scot.butzbach.net/redpack/d/$pkg-x86_64.tar.lz4 ;;
		python-[a-z]*)
			url=$(grep -E "^$pkg\|" /var/lib/redpack/urls.python.redpack | awk -F '|' '{print $2}')
			download $url
			download https://scot.butzbach.net/redpack/d/$pkg-install.sh
			;;
		*) download https://scot.butzbach.net/redpack/d/$pkg-x86_64.tar.lz4 ;;
	esac
done < /var/lib/redpack/list.installed.redpack
echo " -- Verifying packages..."
while read pkg; do
	grep -E "$pkg" /var/lib/redpack/hash.master.redpack > hash.$pkg.redpack
	sha256sum -c --status hash.$pkg.redpack
	if [ $? -ne 0 ]; then
		echo "\"$(grep -oE $pkg.* hash.$pkg.redpack)\" has a bad checksum. Exiting..."
		popd > /dev/null
		rm -r /tmp/redpack
		exit 1
	fi
	echo "$(grep -oE $pkg.* hash.$pkg.redpack): OK."
done < /var/lib/redpack/list.installed.redpack
echo "All packages are verified."
echo ""

mkdir /var/lib/redpack/uninstall
while read pkg; do
	echo -n " -- Installing $pkg..."
	case $pkg in
		python-doc-*)
			mkdir $pkg
			lz4 -d $pkg-x86_64.tar.lz4 - | tar -xpC $pkg --same-owner
			pushd $pkg > /dev/null
			mv uninstall.sh /var/lib/redpack/uninstall/$pkg-uninstall.sh
			OIFS="$IFS"
			IFS=$'\n'
			for x in $(find . | sort); do
				case $(stat -c %F "$x") in
					"directory") install -d "/$x" -m $(stat -c %a "$x") -o $(stat -c %U "$x") -g $(stat -c %G "$x") ;;
					"symbolic link") ln -sf $(readlink "$x") "/$x" ;;
					"regular file") install "$x" "/$x" -m $(stat -c %a "$x") -o $(stat -c %U "$x") -g $(stat -c %G "$x") ;;
				esac
			done
			IFS="$OIFS"
			popd > /dev/null
			rm -rf $pkg
			;;
		perl-[a-z]*|python-[a-z]*) bash $pkg-install.sh 1>/dev/null 2>&1 ;;
		*)
			mkdir $pkg
			lz4 -d $pkg-x86_64.tar.lz4 - | tar -xpC $pkg --same-owner
			pushd $pkg > /dev/null
			if [ -f uninstall.sh ]; then mv uninstall.sh /var/lib/redpack/uninstall-$pkg.sh; fi
			OIFS="$IFS"
			IFS=$'\n'
			for x in $(find . | sort); do
				case $(stat -c %F "$x") in
					"directory") install -d "/$x" -m $(stat -c %a "$x") -o $(stat -c %U "$x") -g $(stat -c %G "$x") ;;
					"symbolic link") ln -sf $(readlink "$x") "/$x" ;;
					"regular file") install "$x" "/$x" -m $(stat -c %a "$x") -o $(stat -c %U "$x") -g $(stat -c %G "$x") ;;
				esac
			done
			IFS="$OIFS"
			popd > /dev/null
			rm -rf $pkg
			;;
	esac
	echo " Done."
done < /var/lib/redpack/list.installed.redpack

# install default configuration
echo ""
echo " -- Fetching configuration tarball..."
download https://scot.butzbach.net/redpack/d/redclaw-default-config-1_rc.2.tar.lz4

echo ""
echo -n " -- Verifying tarball..."
grep -E "redclaw-default-config-1_rc.2" /var/lib/redpack/hash.master.redpack > hash.redclaw-default-config-1_rc.2.redpack
sha256sum -c hash.redclaw-default-config-1_rc.2.redpack
if [ $? -ne 0 ]; then
	echo " Bad checksum. Exiting..."
	popd > /dev/null
	rm -r /tmp/redpack
	exit 1
fi
echo " OK."

echo ""
echo -n " -- Installing configuration files..."
lz4 -d redclaw-default-config-1_rc.2.tar.lz4 - | tar -xp --same-owner
pushd redclaw-default-config-1_rc.2 > /dev/null
for x in aliases fuse.conf; do install {config,/etc}/$x -m $(stat -c %a config/$x) -o $(stat -c %U config/$x) -g $(stat -c %G config/$x); done
for x in {rfcomm,uart}.conf; do install {config,/etc/bluetooth}/$x -m $(stat -c %a config/$x) -o $(stat -c %U config/$x) -g $(stat -c %G config/$x); done
for x in atd ch{age,passwd} exim login other passwd su sudo system{-{auth,account,password,session},d-user}; do install {config,/etc/pam.d}/$x -m $(stat -c %a config/$x) -o $(stat -c %U config/$x) -g $(stat -c %G config/$x); done
install {config,/etc/dbus-1}/session-local.conf -m $(stat -c %a config/session-local.conf) -o $(stat -c %U config/session-local.conf) -g $(stat -c %G config/session-local.conf)
install {config,/etc/sudoers.d}/00-sudo -m $(stat -c %a config/00-sudo) -o $(stat -c %U config/00-sudo) -g $(stat -c %G config/00-sudo)
popd > /dev/null
rm -rf redclaw-default-config-1_rc.2
echo " Done."

popd > /dev/null
rm -r /tmp/redpack

# configure packages
echo ""
echo " -- Enabling Systemd units..."
systemctl daemon-reload
systemctl enable atd
systemctl enable bluetooth
systemctl enable exim
systemctl enable fcron
systemctl enable gpm
systemctl enable iptables
systemctl enable --global obex
systemctl enable smartd
systemctl enable sshd
systemctl enable update-pki.timer

echo ""
echo " -- Creating symbolic links..."
ln -sf exim /usr/sbin/sendmail
if [ $INSTALLDOCS -eq 1 ]; then
	ln -sfn python-3.12.5 /usr/share/doc/python-3
	ln -sf ../../doc/vim-9.1.0660 /usr/share/vim/vim91/doc
fi

echo ""
echo " -- Updating library reference..."
mv /usr/lib/libnssckbi.so{,.orig}
ln -sf pkcs11/p11-kit-trust.so /usr/lib/libnssckbi.so

echo ""
echo " -- Updating XML catalog..."
xmlcatalog --noout --create /etc/xml/docbook
xmlcatalog --noout --create /etc/xml/catalog

xmlcatalog --noout --add "public" "-//OASIS//DTD DocBook XML V4.5//EN" "http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd" /etc/xml/docbook
xmlcatalog --noout --add "public" "-//OASIS//DTD DocBook XML CALS Table Model V4.5//EN" "file:///usr/share/xml/docbook/xml-dtd-4.5/calstblx.dtd" /etc/xml/docbook
xmlcatalog --noout --add "public" "-//OASIS//DTD XML Exchange Table Model 19990315//EN" "file:///usr/share/xml/docbook/xml-dtd-4.5/soextblx.dtd" /etc/xml/docbook
xmlcatalog --noout --add "public" "-//OASIS//ELEMENTS DocBook XML Information Pool V4.5//EN" "file:///usr/share/xml/docbook/xml-dtd-4.5/dbpoolx.mod" /etc/xml/docbook
xmlcatalog --noout --add "public" "-//OASIS//ELEMENTS DocBook XML Document Hierarchy V4.5//EN" "file:///usr/share/xml/docbook/xml-dtd-4.5/dbhierx.mod" /etc/xml/docbook
xmlcatalog --noout --add "public" "-//OASIS//ELEMENTS DocBook XML HTML Tables V4.5//EN" "file:///usr/share/xml/docbook/xml-dtd-4.5/htmltblx.mod" /etc/xml/docbook
xmlcatalog --noout --add "public" "-//OASIS//ENTITIES DocBook XML Notations V4.5//EN" "file:///usr/share/xml/docbook/xml-dtd-4.5/dbnotnx.mod" /etc/xml/docbook
xmlcatalog --noout --add "public" "-//OASIS//ENTITIES DocBook XML Character Entities V4.5//EN" "file:///usr/share/xml/docbook/xml-dtd-4.5/dbcentx.mod" /etc/xml/docbook
xmlcatalog --noout --add "public" "-//OASIS//ENTITIES DocBook XML Additional General Entities V4.5//EN" "file:///usr/share/xml/docbook/xml-dtd-4.5/dbgenent.mod" /etc/xml/docbook
xmlcatalog --noout --add "rewriteSystem" "http://www.oasis-open.org/docbook/xml/4.5" "file:///usr/share/xml/docbook/xml-dtd-4.5" /etc/xml/docbook
xmlcatalog --noout --add "rewriteURI" "http://www.oasis-open.org/docbook/xml/4.5" "file:///usr/share/xml/docbook/xml-dtd-4.5" /etc/xml/docbook
xmlcatalog --noout --add "delegatePublic" "-//OASIS//ENTITIES DocBook XML" "file:///etc/xml/docbook" /etc/xml/catalog
xmlcatalog --noout --add "delegatePublic" "-//OASIS//DTD DocBook XML" "file:///etc/xml/docbook" /etc/xml/catalog
xmlcatalog --noout --add "delegateSystem" "http://www.oasis-open.org/docbook/" "file:///etc/xml/docbook" /etc/xml/catalog
xmlcatalog --noout --add "delegateURI" "http://www.oasis-open.org/docbook/" "file:///etc/xml/docbook" /etc/xml/catalog

for DTDVERSION in 4.1.2 4.{2..4}; do
	xmlcatalog --noout --add "public" "-//OASIS//DTD DocBook XML V$DTDVERSION//EN" "http://www.oasis-open.org/docbook/xml/$DTDVERSION/docbookx.dtd" /etc/xml/docbook
	xmlcatalog --noout --add "rewriteSystem" "http://www.oasis-open.org/docbook/xml/$DTDVERSION" "file:///usr/share/xml/docbook/xml-dtd-4.5" /etc/xml/docbook
	xmlcatalog --noout --add "rewriteURI" "http://www.oasis-open.org/docbook/xml/$DTDVERSION" "file:///usr/share/xml/docbook/xml-dtd-4.5" /etc/xml/docbook
	xmlcatalog --noout --add "delegateSystem" "http://www.oasis-open.org/docbook/xml/$DTDVERSION/" "file:///etc/xml/docbook" /etc/xml/catalog
	xmlcatalog --noout --add "delegateURI" "http://www.oasis-open.org/docbook/xml/$DTDVERSION/" "file:///etc/xml/docbook" /etc/xml/catalog
done

xmlcatalog --noout --add "rewriteSystem" "http://cdn.docbook.org/release/xsl-nons/1.79.2" "file:///usr/share/xml/docbook/xsl-stylesheets-nons-1.79.2" /etc/xml/catalog
xmlcatalog --noout --add "rewriteSystem" "https://cdn.docbook.org/release/xsl-nons/1.79.2" "file:///usr/share/xml/docbook/xsl-stylesheets-nons-1.79.2" /etc/xml/catalog
xmlcatalog --noout --add "rewriteURI" "http://cdn.docbook.org/release/xsl-nons/1.79.2" "file:///usr/share/xml/docbook/xsl-stylesheets-nons-1.79.2" /etc/xml/catalog
xmlcatalog --noout --add "rewriteURI" "https://cdn.docbook.org/release/xsl-nons/1.79.2" "file:///usr/share/xml/docbook/xsl-stylesheets-nons-1.79.2" /etc/xml/catalog
xmlcatalog --noout --add "rewriteSystem" "http://cdn.docbook.org/release/xsl-nons/current" "file:///usr/share/xml/docbook/xsl-stylesheets-nons-1.79.2" /etc/xml/catalog
xmlcatalog --noout --add "rewriteSystem" "https://cdn.docbook.org/release/xsl-nons/current" "file:///usr/share/xml/docbook/xsl-stylesheets-nons-1.79.2" /etc/xml/catalog
xmlcatalog --noout --add "rewriteURI" "http://cdn.docbook.org/release/xsl-nons/current" "file:///usr/share/xml/docbook/xsl-stylesheets-nons-1.79.2" /etc/xml/catalog
xmlcatalog --noout --add "rewriteURI" "https://cdn.docbook.org/release/xsl-nons/current" "file:///usr/share/xml/docbook/xsl-stylesheets-nons-1.79.2" /etc/xml/catalog
xmlcatalog --noout --add "rewriteSystem" "http://docbook.sourceforge.net/release/xsl/current" "file:///usr/share/xml/docbook/xsl-stylesheets-nons-1.79.2" /etc/xml/catalog
xmlcatalog --noout --add "rewriteURI" "http://docbook.sourceforge.net/release/xsl/current" "file:///usr/share/xml/docbook/xsl-stylesheets-nons-1.79.2" /etc/xml/catalog

echo ""
echo " -- Configuring Cracklib..."
create-cracklib-dict /usr/share/dict/*

echo ""
echo " -- Configuring Fcron..."
systemctl start fcron
fcrontab -zu systab

echo ""
echo " -- Generating SSH host keys..."
ssh-keygen -A

echo ""
echo " -- Adding certificates to NSS system database..."
update-ca-certificates

echo ""
echo " -- Updating MIME info..."
update-mime-database /usr/share/mime

echo ""
echo " -- Configuring Shadow..."
pwconv
grpconv
useradd -D --gid 999
mv /etc/login.access{,.orig}
mv /etc/limits{,.orig}

# set root password and create new account (if wanted; also setting a password)
echo ""
echo " -- Setting root password..."
passwd root

# configure firewall
echo ""
echo " -- Configuring iptables..."
echo "Select a firewall - [p]ersonal or [n]etwork"
LOOP=1
while [ $LOOP -eq 1 ]; do
    echo ""
    read -p "Firewall: "
    case $REPLY in
        n)
            echo "A network interface connected to the Internet must be specified to filter traffic going through it."
            echo ""
            echo "Available interfaces:"
            ls /sys/class/net | grep -oE "[e|w].*"
            while [ $LOOP -eq 1 ]; do
                echo ""
                read -p "Interface: "
                for net in $(ls /sys/class/net | grep -oE "[e|w].*"); do
                    if [ $REPLY = $net ]; then
                        echo " -- Using network firewall, filtering traffic going through \"$net\"..."
                        sed "/^WAN1/s/\".*\"/\"$net\"/" -i /etc/systemd/scripts/iptables-network
                        ln -sf iptables-network /etc/systemd/scripts/iptables
                        LOOP=0
                    fi
                done
            done
            ;;
        p)
            echo " -- Using personal firewall..."
            ln -sf iptables-personal /etc/systemd/scripts/iptables
            LOOP=0
            ;;
        *) ;;
    esac
done

echo ""
echo "Make sure to edit /etc/systemd/scripts/iptables to further configure your firewall."
echo "Run \"switch-firewall\" to switch your firewall later."
echo ""

echo " -- Starting iptables..."
systemctl start iptables

echo ""
echo "This may take a while."
mkinitramfs 6.10.5
mv initrd.img-6.10.5 /boot

echo ""
echo "Redclaw Linux is now fully installed and configured on your system."
echo ""
echo "To configure the system later on, refer to the Configuration Guide."
echo ""
echo "To make changes go into effect, \"reboot\" your system."
echo ""
echo "Thank you for choosing Redclaw Linux."
echo ""
